// Copyright © 2015, 2016 T-semi
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef INCLUDED_FINDER_FINDER_H
#define INCLUDED_FINDER_FINDER_H

//コメントアウト解除するとちょっと軽くなるかも。ON推奨
#define FINDER_LIGHT_APIX

#include <stdint.h>

#include <string>
#include <vector>
#ifdef FINDER_GET_EXECUTION_TIME
    #include <iostream>
#endif

namespace finder{
//定数
extern const int light_threshold;
extern const int stage1_allowed_distfluc;
extern const double stage1_allowed_distfluc_rate;

extern const int stage2_ignoreable;
extern const double stage2_ignoreable_rate;
extern const int stage2_allowed_distfluc;
extern const double stage2_allowed_distfluc_rate;

extern const double merge_radius_fluc_rate;
extern const double merge_distance_rate;

extern const int circletest_ignoreable;
extern const double circletest_ignoreable_rate;
extern const int circletest_allowed_distfluc;
extern const double circletest_allowed_distfluc_rate;

extern const double direction_accuracy;
extern const int direction_marker_maxnum;
extern const int direction_marker_w;
extern const int direction_marker_w_dmax;
extern const int direction_marker_w_min;
extern const int direction_marker_w_max;

struct apix;
struct marker;
class finder_context;

//各ピクセルの情報を格納するための構造体
struct apix{
#ifndef FINDER_LIGHT_APIX
    //フラグのやつ
    static const int nothing;
    static const int scan_high2low;
    static const int scan_low2high;
    static const int possive_center;
    static const int stripe_testp;
    static const int droped;
    static const int asstripe;
    static const int bycircletest;
    static const int distance_line;

    int flags;

    struct{uint8_t r; uint8_t g; uint8_t b;} color;
#endif

    int gscale;             //最大値は0xffではなく0xff*3
};

//何かしらの位置
struct point2d{
    double x;
    double y;
};

//マーカーの位置とかいろいろを格納するはず。
struct marker: point2d{
    double r;
    double gray;
    int gray_int;
    double direction;
    int id;
};

//画像中の座標->現実の位置 に変換するclass(基底class)
class axistrans{
public:
    virtual ~axistrans() = 0;

    virtual point2d transform(point2d &origin);
    virtual marker transform(marker &origin) = 0;
};

//何もしないで元の値を返すaxistrans
class no_axistrans: public axistrans{
public:
    no_axistrans();
    ~no_axistrans();

    point2d transform(point2d &origin);
    marker transform(marker &origin);
};

//シンプルなaxistrans。画像中の座標2箇所と、それに対応する座標2箇所を指定する
//うまく動作しないような気がします。
class simple_axistrans: public axistrans{
private:
    double x_rate;
    double y_rate;
    double x_diff;
    double y_diff;
    double xy_rate;
    double rate;

public:
    simple_axistrans(point2d from1, point2d from2, point2d to1, point2d to2);
    ~simple_axistrans();

    point2d transform(point2d &origin);
    marker transform(marker &origin);
};

//マーカーを探すヤツ。フレームの大きさを与えてコンテキスト化した後は、
//set_frameでフレームをセットし、find_allを呼ぶことでマーカーを見つけてくれる。
//結果はmarkersに代入される。
//set_axistransでaxistransを設定すると、get_markersで変換できる。
//set_axistransは、ポインタを渡した場合、不要になった時にaxistransをdeleteする。
//参照を渡した場合、axistransのコンテキストは呼出元が管理する(コピーされない)。
class finder_context{
private:
    bool have_atrans;
    axistrans *atrans;

public:
    finder_context(int _w, int _h);
    virtual ~finder_context();

    std::vector< std::vector<marker> > markers_prev;
    std::vector<marker> markers;
    apix *bitmap;
    const int w, h;

    void set_frame(const apix *fdat);
    void set_frame(const std::string &rgb_string);

    virtual void track_all();
    virtual void find_all();

    void set_axistrans(axistrans &context);
    void set_axistrans(axistrans *context);
    std::vector<marker> get_markers();
    void get_markers(std::vector<marker> &ret);

    size_t get_marker_num();
};

void mark_possibility(apix *fdat, int w, int h, std::vector<marker> &markers);
void drop_stripe(apix *fdat, int w, int h, std::vector<marker> &markers);
void merge_markers(apix *fdat, int w, int h, std::vector<marker> &markers);
void snap_markers(apix *fdat, int w, int h, std::vector<marker> &markers);
void get_marker_gray(apix *fdat, int w, int h, std::vector<marker> &markers);
void marker_test_circle(apix *fdat, int w, int h, std::vector<marker> &markers);
void decide_direction(apix *fdat, int w, int h, std::vector<marker> &markers);

void add_tracking_point
        (std::vector<marker> &markers,
         const std::vector< std::vector<marker> > &markers_prev);
}

#endif  //INCLUDED_FINDER_FINDER_H
