// Copyright © 2015, 2016 T-semi
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "finder_adv.h"

#include <algorithm>
#include <cmath>
#include <climits>

//#define FINDER_ADV_DEBUG
#ifdef FINDER_ADV_DEBUG
#include <iostream>
#endif

using namespace finder;

const int finder_adv::id_overwrite_threash = 20;
const int finder_adv::marker_remove_threash = 5;

finder_adv::finder_adv(int w, int h, int num_markers):
finder_context(w, h), markers_required(num_markers),
not_enough_count(0), id_unknown_count(0){
}

finder_adv::~finder_adv(){
}

void finder_adv::find_all(){
    bool idunknown_flag = false;

    finder_context::find_all();
    last_decide.clear();
    std::vector<marker_adv> &new_decide = last_decide;

    not_enough_count = 0;
    id_unknown_count = 0;

    while(markers.size() && new_decide.size() < markers_required){
        marker_adv ma;
        static_cast<marker &>(ma) = markers.front();
        ma.notdetect_count = 0;
        if(ma.id){
            ma.iderror_count = 0;
        }else{
            ma.iderror_count = id_overwrite_threash;
            idunknown_flag = true;
        }
        new_decide.push_back(ma);
        markers.erase(markers.begin());
    }

    if(new_decide.size() != markers_required){
        ++not_enough_count;
    }
    if(idunknown_flag){
        ++id_unknown_count;
    }

    markers.clear();
    for(const marker_adv &mp: new_decide){
        markers.push_back(static_cast<marker>(mp));
    }
}

void finder_adv::track_all(){
    bool idunknown_flag = false;

    finder_context::track_all();

    std::vector<marker_adv> &new_decide = last_decide;

    for(marker_adv &ma: new_decide){
        int mindist = ma.r;
        int efid = -1;
        int head = 0;
        for(const marker &dm: markers){
            double dx = dm.x - ma.x;
            double dy = dm.y - ma.y;
            double dist = std::sqrt((dx * dx) + (dy * dy));
            if(dist < mindist){
                mindist = dist;
                efid = head;
            }
            ++head;
        }
        if(efid >= 0){
            std::vector<marker>::iterator p = markers.begin() + efid;
            if(p->id){
                if(p->id != ma.id){
                    if(++ma.iderror_count > id_overwrite_threash){
                        ma.direction = p->direction;
                        ma.id = p->id;
                    }
                }else{
                    ma.direction = p->direction;
                    ma.iderror_count = 0;
                }
            }else{
#ifdef FINDER_ADV_DEBUG
                std::cout << "id estimated" << std::endl;
#endif
                ++ma.iderror_count;
                if(ma.iderror_count >= INT_MAX){
                    ma.iderror_count = INT_MAX - 1;
                }
                idunknown_flag = true;
            }
            ma.x = p->x;
            ma.y = p->y;
            ma.r = p->r;
            ma.gray = p->gray;
            ma.gray_int = p->gray_int;

            markers.erase(p);
            ma.notdetect_count = 0;
        }else{
#ifdef FINDER_ADV_DEBUG
                std::cout << "pos estimated" << std::endl;
#endif
            ++ma.notdetect_count;
            if(ma.notdetect_count >= INT_MAX){
                ma.notdetect_count = INT_MAX - 1;
            }
        }
    }

    if(new_decide.size() != markers_required){
        if(not_enough_count < INT_MAX){
            ++not_enough_count;
        }
    }else{
        not_enough_count = 0;
    }

    if(idunknown_flag){
        if(id_unknown_count < INT_MAX){
            ++id_unknown_count;
        }
    }else{
        id_unknown_count = 0;
    }

    std::vector<marker_adv>::iterator p = new_decide.begin();
    while(p < new_decide.end()){
        if(p->notdetect_count >= marker_remove_threash){
#ifdef FINDER_ADV_DEBUG
            std::cout << "mdec" << std::endl;
#endif
            p = new_decide.erase(p);
        }else{
            ++p;
        }
    }

#ifdef FINDER_ADV_DEBUG
    std::cout << "fml:" << markers.size() << std::endl;
    std::cout << "tlm:" << new_decide.size() << std::endl;
#endif
    while(markers.size() && new_decide.size() < markers_required){
#ifdef FINDER_ADV_DEBUG
        std::cout << "minc" << std::endl;
#endif
        marker_adv ma;
        static_cast<marker &>(ma) = markers.front();
        ma.notdetect_count = 0;
        if(ma.id){
            ma.iderror_count = 0;
        }else{
            ma.iderror_count = id_overwrite_threash;
        }
        new_decide.push_back(ma);
        markers.erase(markers.begin());
    }

    markers.clear();
    for(const marker_adv &mp: new_decide){
        markers.push_back(static_cast<marker>(mp));
    }
}
