// Copyright © 2015, 2016 T-semi
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

//コメントアウト解除で、stdoutに実行時間についての情報を出力
//#define GET_RUN_TIME

#include <stdint.h>
#include <time.h>

#include <cmath>
#include <iostream>
#include <vector>

#include <SDL.h>

#ifdef GET_RUN_TIME
    #include "stopwatch.h"
#endif

#include "v4l2.h"
#include "finder_adv.h"

//定数
//const int frame_width = 640;
//const int frame_height = 480;
const int frame_width = 1280;
const int frame_height = 720;
//const int frame_width = 1920;
//const int frame_height = 1080;
const int camera_fps = 20;

//グローバルな変数
v4l2::camera *the_camera;
finder::finder_adv *fc;

void init(){
    the_camera = new v4l2::camera("/dev/video0");

    the_camera->open();

    the_camera->set_format(V4L2_PIX_FMT_YUYV);
    the_camera->set_framesize(frame_width, frame_height);
    the_camera->set_frame_interval(1, camera_fps);

    the_camera->init();

    fc = new finder::finder_adv(frame_width, frame_height, 4);
}

void wait_cam(){
    the_camera->wait_frame(1, 0);
}

//こいつが各フレームの絵を作る
std::string get_frame(){
    std::string camera_frame = the_camera->get_rgbframe_string();

    if(camera_frame.size() != frame_width * frame_height * 3){
        //画像のデータサイズが規格外なら空データを返す
        return std::string();
    }

#ifdef GET_RUN_TIME
    stopwatch::watch wins;
    wins.plot();
#endif

    fc->set_frame(camera_frame);
#ifdef GET_RUN_TIME
    wins.plot();
#endif
    fc->track_all();
#ifdef GET_RUN_TIME
    wins.plot();
#endif
    std::vector<finder::marker> markers = fc->get_markers();
    finder::apix *fdat = fc->bitmap;

#ifdef GET_RUN_TIME
    wins.plot();
    wins.print();
#endif

    //描画用データ生成
    std::string ret(frame_width * frame_height * 3, 0);
    for(size_t i = 0; i < frame_width * frame_height; ++i){
#ifdef FINDER_LIGHT_APIX
        ret[i*3]     = fdat[i].gscale / 3;
        ret[i*3 + 1] = ret[i*3];
        ret[i*3 + 2] = ret[i*3];
#else
        if(fdat[i].flags == finder::apix::nothing){
            ret[i*3]     = fdat[i].gscale / 3;
            ret[i*3 + 1] = ret[i*3];
            ret[i*3 + 2] = ret[i*3];
        }else if(fdat[i].flags & finder::apix::scan_low2high){
            ret[i*3]     = 0x88;
            ret[i*3 + 1] = 0x00;
            ret[i*3 + 2] = 0x00;
        }else if(fdat[i].flags & finder::apix::scan_high2low){
            ret[i*3]     = 0x00;
            ret[i*3 + 1] = 0x00;
            ret[i*3 + 2] = 0x88;
        }else if(fdat[i].flags & finder::apix::distance_line){
            ret[i*3]     = 0xff;
            ret[i*3 + 1] = 0xff;
            ret[i*3 + 2] = 0xff;
        }else if(fdat[i].flags & finder::apix::droped){
            if(fdat[i].flags & finder::apix::asstripe){
                ret[i*3]     = 0x00;
                ret[i*3 + 1] = 0x88;
                ret[i*3 + 2] = 0x88;
            }else if(fdat[i].flags & finder::apix::bycircletest){
                ret[i*3]     = 0x00;
                ret[i*3 + 1] = 0xff;
                ret[i*3 + 2] = 0x00;
            }else{
                ret[i*3]     = 0xff;
                ret[i*3 + 1] = 0x00;
                ret[i*3 + 2] = 0x00;
            }
        }else if(fdat[i].flags & finder::apix::possive_center){
            ret[i*3]     = 0x00;
            ret[i*3 + 1] = 0xff;
            ret[i*3 + 2] = 0xff;
        }else if(fdat[i].flags & finder::apix::stripe_testp){
            ret[i*3]     = 0x88;
            ret[i*3 + 1] = 0x88;
            ret[i*3 + 2] = 0x00;
        }else{
            ret[i*3]     = fdat[i].gscale / 3;
            ret[i*3 + 1] = ret[i*3];
            ret[i*3 + 2] = ret[i*3];
        }
#endif
    }

    std::cout << "detected: " << markers.size() << std::endl;
    std::cout << "ID unknown: " << fc->id_unknown_count << std::endl;
    std::cout << "NE Markers: " << fc->not_enough_count << std::endl;

    std::vector<finder::marker>::iterator p = markers.begin();
    std::vector<finder::marker>::iterator end = markers.end();
    for(p; p != end; ++p){
        double r2 = p->r * p->r;
        for(int dx = - p->r; dx < p->r; ++dx){
            int x = p->x + dx + 0.5;
            int y1 = std::sqrt(r2 - dx * dx) + p->y + 0.5;
            int y2 = -std::sqrt(r2 - dx * dx) + p->y + 0.5;
            int i1 = (y1 * frame_width + x) * 3;
            int i2 = (y2 * frame_width + x) * 3;
            ret[i1]      = 0xff;
            ret[i1 + 1]  = 0x00;
            ret[i1 + 2]  = 0xff;
            ret[i2]      = 0xff;
            ret[i2 + 1]  = 0x00;
            ret[i2 + 2]  = 0xff;
        }
        if(p->id){
            int start_x = p->x + 0.5;
            int start_y = p->y + 0.5;
            int end_x = p->x + (p->r) * std::cos(p->direction) + 0.5;
            int end_y = p->y + (p->r) * std::sin(p->direction) + 0.5;

            int xsteps = end_x - start_x;
            int ysteps = end_y - start_y;
            int xsteps_p = xsteps<0 ? -xsteps : xsteps;
            int ysteps_p = ysteps<0 ? -ysteps : ysteps;
            double steps = xsteps_p;
            if(steps < ysteps_p){
                steps = ysteps_p;
            }
            for(int i = 0; i < steps; ++i){
                int x = start_x + (xsteps / steps) * i + 0.5;
                int y = start_y + (ysteps / steps) * i + 0.5;
                int dp = (y*frame_width + x) * 3;
                ret[dp]      = 0xff;
                ret[dp + 1]  = 0x00;
                ret[dp + 2]  = 0x00;
            }
        }
    }

    return ret;
}

void free_cam(){
    the_camera->uninit();
    the_camera->close();
    delete the_camera;
}

//main関数は↑で定義した関数を使ってgetしたフレームをSDLで描きまくるだけ
int main(){
    int i;

    init();

    SDL_Window *window;                    // ポインタを宣言する
    SDL_Renderer *renderer;

    if(SDL_Init(SDL_INIT_VIDEO) < 0){              // SDL2を初期化する
        //エラー処理
        return 1;
    }

    // 次の設定でアプリケーションウィンドウを生成する:
    window = SDL_CreateWindow(
        "An SDL2 window",                  // ウィンドウのタイトル
        SDL_WINDOWPOS_UNDEFINED,           // X座標の初期値
        SDL_WINDOWPOS_UNDEFINED,           // Y座標の初期値
        frame_width,                       // 幅のピクセル数
        frame_height,                      // 高さのピクセル数
        SDL_WINDOW_RESIZABLE               // フラグ
    );

    // ウィンドウの生成に成功したかチェックする
    if (window == NULL) {
        // ウィンドウを生成できなかった...
        printf("ウィンドウを生成できなかった: %s\n", SDL_GetError());
        return 1;
    }

//    renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_SOFTWARE);
    renderer = SDL_CreateRenderer(window, -1, 0);
    SDL_SetRenderDrawColor(renderer, 255, 255, 255, 255);
    SDL_RenderClear(renderer);
    SDL_RenderPresent(renderer);

    uint32_t window_id = SDL_GetWindowID(window);

    time_t last_time = 0;
    unsigned int numsamples = 0;
    while(true){
        SDL_Event event;
        while(SDL_PollEvent(&event)){
            if(event.type == SDL_WINDOWEVENT){
                if(event.window.event == SDL_WINDOWEVENT_CLOSE &&
                    event.window.windowID == window_id){
                    SDL_DestroyWindow(window);
                }
            }else if(event.type == SDL_QUIT){
                goto exit_main_loop;
            }else if(event.type == SDL_KEYUP){
                if(event.key.keysym.sym == SDLK_q){
                    SDL_DestroyWindow(window);
                    goto exit_main_loop;
                }
            }
        }
        int w, h;
        SDL_GetRendererOutputSize(renderer, &w, &h);
        SDL_SetRenderDrawColor(renderer, 255, 255, 255, 255);
        SDL_RenderClear(renderer);

        SDL_Texture *stex;
        wait_cam();
        std::string rst = get_frame();
        if(rst.size()){
            uint32_t *pixels = static_cast<uint32_t *>(malloc(sizeof(uint32_t) * frame_width * frame_height));
            int x, y;
            for(y = 0; y < frame_height; y++){
                for(x = 0; x < frame_width; x++){
                    size_t basepos = (y * frame_width + x)*3;
                    uint32_t bl = uint32_t(uint8_t(rst[basepos]))
                                | (uint32_t(uint8_t(rst[basepos+1])) << 8)
                                | (uint32_t(uint8_t(rst[basepos+2])) << 16);
                    pixels[x + y * frame_width] = bl | 0xff000000;
                }
            }
            SDL_Surface *ssf = SDL_CreateRGBSurfaceFrom(pixels,
                frame_width, frame_height, 32, frame_width * sizeof(uint32_t),
                0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000);
            if(ssf == NULL){
                fprintf(stderr, "creation suf fail: %s\n", SDL_GetError());
                exit(1);
            }
            stex = SDL_CreateTextureFromSurface(renderer, ssf);
            if(stex == NULL){
                fprintf(stderr, "creation tex fail: %s\n", SDL_GetError());
                exit(1);
            }
            SDL_FreeSurface(ssf);
            free(pixels);
            SDL_RenderCopy(renderer, stex, NULL, NULL);
            SDL_RenderPresent(renderer);
            SDL_DestroyTexture(stex);

            time_t time_now = time(NULL);
            if(last_time + 3 <= time_now){
                std::cout << numsamples / (time_now - last_time)
                          << "fps" << std::endl;
                last_time = time_now;
                numsamples = 0;
            }
            ++numsamples;
        }
    }
exit_main_loop:

    SDL_Quit();

    free_cam();

    return 0;
}
