// Copyright © 2015, 2016 T-semi
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef INCLUDED_STOPWATCH_STOPWATCH_H
#define INCLUDED_STOPWATCH_STOPWATCH_H

#include <time.h>
#include <vector>
#include <iostream>
#include <stdexcept>

//主に実行時間を計測するためのもの。
//最適化するために時間を測りたいときに使う小道具
namespace stopwatch{
    class timespec_adv;
    class watch;

    //timespecが、使いにくいので適当に改造
    class timespec_adv: public timespec{
    public:
        double to_sec();
        timespec_adv operator-(timespec &v);
    };

    //plot()を呼ぶと時刻を記録し、print()を呼ぶと記録をstdoutに吐くだけ
    class watch{
    public:
        std::vector<timespec_adv> plots;
        void plot();
        void print();
    };

    inline double timespec_adv::to_sec(){
        return tv_sec + tv_nsec / 1000000000.0;
    }

    inline timespec_adv timespec_adv::operator-(timespec &v){
        timespec_adv r;
        r.tv_sec = this->tv_sec - v.tv_sec;
        r.tv_nsec = this->tv_nsec - v.tv_nsec;
        if(r.tv_nsec < 0){
            --r.tv_sec;
            r.tv_nsec += 1000000000;
        }
        return r;
    }

    inline void watch::plot(){
        timespec_adv time;
        if(clock_gettime(CLOCK_MONOTONIC, &time)){
            throw std::runtime_error("clock_gettime error");
        }
        plots.push_back(time);
    }

    inline void watch::print(){
        size_t size = plots.size();
        for(size_t i = 0; i < size; ++i){
            std::cout << "[" << i << "]"
                      << plots[i].to_sec() << "sec ([0]+"
                      << (plots[i]-plots[0]).to_sec() << "sec";
            if(i){
                std::cout << ", [" << i-1 << "]+"
                          << (plots[i]-plots[i-1]).to_sec() << "sec";
            }
            std::cout << ")" << std::endl;
        }
    }
}
#endif  //INCLUDED_STOPWATCH_STOPWATCH_H
