// Copyright © 2015, 2016 T-semi
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <stdint.h>
#include <time.h>

#include <iostream>
#include <vector>

#include <SDL.h>

#include "v4l2.h"

int main(){
    int i;
    v4l2::camera c("/dev/video0");
    c.open();
    i = 0;
    while(true){
        try{
            v4l2_fmtdesc f = c.get_format(i);
            std::cout << f.description;
            std::cout << "(ID:0x"
                      << std::hex << f.pixelformat << std::dec
                      << ")" << std::endl;
            int j = 0;
            while(true){
                try{
                    v4l2_frmsizeenum s = c.get_frame_size(f.pixelformat, j);
                    if(s.type == V4L2_FRMSIZE_TYPE_DISCRETE){
                        std::cout << s.discrete.width << "x"
                                  << s.discrete.height << "\t";
                        int k = 0;
                        while(true){
                            try{
                                v4l2_frmivalenum i = c.get_frame_interval(
                                        f.pixelformat,
                                        s.discrete.width, s.discrete.height, k);
                                if(i.type == V4L2_FRMIVAL_TYPE_DISCRETE){
                                    std::cout << i.discrete.numerator << "/"
                                              << i.discrete.denominator << " ";
                                }
                            }catch(std::invalid_argument){
                                break;
                            }
                            ++k;
                        }
                        std::cout << std::endl;
                    }
                }catch(std::invalid_argument){
                    break;
                }
                ++j;
            }
        }catch(std::out_of_range){
            break;
        }
        ++i;
    }
#ifdef PIX_MJPEG
    c.set_format(V4L2_PIX_FMT_MJPEG);
    std::cout << "pixfmt: mjpeg" << std::endl;
#else
    c.set_format(V4L2_PIX_FMT_YUYV);
    std::cout << "pixfmt: yuyv" << std::endl;
#endif
    c.set_framesize(640, 480);
    c.set_frame_interval(1, 30);
    c.init();

    SDL_Window *window;                    // ポインタを宣言する
    SDL_Renderer *renderer;

    if(SDL_Init(SDL_INIT_VIDEO) < 0){              // SDL2を初期化する
        //エラー処理
        return 1;
    }

    // 次の設定でアプリケーションウィンドウを生成する:
    window = SDL_CreateWindow(
        "An SDL2 window",                  // ウィンドウのタイトル
        SDL_WINDOWPOS_UNDEFINED,           // X座標の初期値
        SDL_WINDOWPOS_UNDEFINED,           // Y座標の初期値
        640,                               // 幅のピクセル数
        480,                               // 高さのピクセル数
        SDL_WINDOW_RESIZABLE                  // フラグ
    );

    // ウィンドウの生成に成功したかチェックする
    if (window == NULL) {
        // ウィンドウを生成できなかった...
        printf("ウィンドウを生成できなかった: %s\n", SDL_GetError());
        return 1;
    }

//    renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_SOFTWARE);
    renderer = SDL_CreateRenderer(window, -1, 0);
    SDL_SetRenderDrawColor(renderer, 255, 255, 255, 255);
    SDL_RenderClear(renderer);
    SDL_RenderPresent(renderer);

    uint32_t window_id = SDL_GetWindowID(window);

    time_t last_time = 0;
    unsigned int numsamples = 0;
    while(true){
        SDL_Event event;
        while(SDL_PollEvent(&event)){
            if(event.type == SDL_WINDOWEVENT){
                if(event.window.event == SDL_WINDOWEVENT_CLOSE &&
                    event.window.windowID == window_id){
                    SDL_DestroyWindow(window);
                }
            }else if(event.type == SDL_QUIT){
                goto exit_main_loop;
            }else if(event.type == SDL_KEYUP){
                if(event.key.keysym.sym == SDLK_q){
                    SDL_DestroyWindow(window);
                    goto exit_main_loop;
                }
            }
        }
        int w, h;
        SDL_GetRendererOutputSize(renderer, &w, &h);
        SDL_SetRenderDrawColor(renderer, 255, 255, 255, 255);
        SDL_RenderClear(renderer);

        SDL_Texture *stex;
        //std::string rst = c.get_rawframe_string();
        c.wait_frame(2, 0);
        std::string rst = c.get_rgbframe_string();
        if(rst.size()){
            uint32_t *pixels = static_cast<uint32_t *>(malloc(sizeof(uint32_t) * 640 * 480));
            int x, y;
            for(y = 0; y < 480; y++){
                for(x = 0; x < 640; x++){
                    size_t basepos = (y * 640 + x)*3;
                    uint32_t bl = uint32_t(uint8_t(rst[basepos]))
                                | (uint32_t(uint8_t(rst[basepos+1])) << 8)
                                | (uint32_t(uint8_t(rst[basepos+2])) << 16);
                    pixels[x + y * 640] = bl | 0xff000000;
                }
            }
            SDL_Surface *ssf = SDL_CreateRGBSurfaceFrom(
                pixels, 640, 480, 32, 640 * sizeof(uint32_t),
                0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000);
            if(ssf == NULL){
                fprintf(stderr, "creation suf fail: %s\n", SDL_GetError());
                exit(1);
            }
            stex = SDL_CreateTextureFromSurface(renderer, ssf);
            if(stex == NULL){
                fprintf(stderr, "creation tex fail: %s\n", SDL_GetError());
                exit(1);
            }
            free(pixels);
            SDL_RenderCopy(renderer, stex, NULL, NULL);
            SDL_RenderPresent(renderer);


            time_t time_now = time(NULL);
            if(last_time + 3 <= time_now){
                std::cout << numsamples / (time_now - last_time)
                          << "fps" << std::endl;
                last_time = time_now;
                numsamples = 0;
            }
            ++numsamples;
        }
    }
exit_main_loop:

    SDL_Quit();

    c.uninit();
    c.close();
    return 0;
}
